-- -*- mode: Lua; indent-tabs-mode: nil; lua-indent-level: 4; -*-

--[[

    Annotations v1.0
    by irons

]]

CustomPersistence = {
    init = function(restoring_game)
	-- taken from Wrk's Istoria main.lua
	if Game.ticks ~= 0 then
	    if restoring_game then
		if not Game.restore_saved() then
		    print("WARNING: restoring from saved state failed")
		end
	    else
		if not Game.restore_passed() then
		    print("WARNING: restoring from passed state failed")
		end
	    end
	end
    end
}

CustomLatches = {
    LATCHES = "_custom_latches_latches",
    HOTKEY_TIME = 3,

    init = function(restoring_game)
	local player, i, latches
	
	for player in Players() do
	    latches = {}
	    player[CustomLatches.LATCHES] = latches
	    
	    for i = 1, 12 do
		latches[i] = {
		    count = 0,
		    started = Game.ticks - 1,
		    stopped = Game.ticks,
		}
	    end
	end
    end,

    idle = function()
	local player
	
	for player in Players() do
	    CustomLatches.register_hotkeys(player)
	end
    end,

    register_hotkeys = function(player)
	local i, keys, key

	keys = player[CustomLatches.LATCHES]
        
	for i = 1, 12 do
	    key = keys[i]
	    
	    if player.hotkey == i then
		key.just_started = key.count == 0
		key.started = Game.ticks
		key.count = CustomLatches.HOTKEY_TIME
	    elseif key.count > 0 then
		key.count = key.count - 1
		key.just_started = false
	    else
		key.stopped = Game.ticks
		key.just_started = false
	    end
	end
    end,

    get_hotkey_state = function(player, hotkey)
	local keys, key
	keys = player[CustomLatches.LATCHES]
	if keys then key = keys[hotkey] end
	return key and key.just_started
    end,

    get_hotkey = function(player)
	local keys, key, i

	keys = player[CustomLatches.LATCHES]
	
	if keys then
	    for i = 1, 12 do
		key = keys[i]
		if key and key.just_started then
		    return i
		end
	    end
	end
	
	return nil
    end,
}

CustomAnnotations = {
    MENU = "_custom_annotations_menu",
    REFRESH = "_custom_annotations_refresh",
    ORIGINAL = "_custom_annotations_original",
    EDITING = "_custom_annotations_editing",
    PERSIST = "_custom_annotations_persisted",

    CUTOFF = 4,
    ELLIPSIS = "...",
    
    NEW = "New",
    EDIT = "Edit",
    DELETE = "Delete",
    SLOT_1 = 1,
    SLOT_2 = 2,
    SLOT_3 = 3,
    SLOT_4 = 4,
    SLOT_5 = 5,
    SLOT_6 = 6,
    DEFAULTS = {
	"[!]",
	"(*)",
	"[A]",
	"[B]",
	"[C]",
        "[D]",
    },
    CONFIRM = "Confirm",
    CANCEL = "Cancel",
    AUTO_RANGE = 3,
    
    TOP_MENU = {}, -- filled in at init
    NEW_MENU = {},
    EDIT_MENU = {},
    DELETE_MENU = {},
    INVALID_ANNOTATION = { invalid = true },
    abbreviations = {},
    flashing = {},
    flashing_sorted = {},
    pool = {},

    init = function(restoring_game)
	local persist, level_persist, idx, annotation, player
	
	persist = Game[CustomAnnotations.PERSIST]

	if not persist then
	    persist = {}
	    Game[CustomAnnotations.PERSIST] = persist
	end

	level_persist = persist[Level.index]

	if not level_persist then
	    level_persist = {}
	    persist[Level.index] = level_persist
	end
	
	for annotation in Annotations() do
	    if restoring_game and annotation[CustomAnnotations.PERSIST] then
		CustomAnnotations.persist(annotation)
	    end

	    last = annotation.index
	end

	local tmp = {}

	for idx, annotation in pairs(level_persist) do
	    if not Annotations[idx] then
		table.insert(tmp, annotation)
	    end
	end

	table.sort(tmp, function(a, b)
		       return a.index < b.index
	end)

	for idx, annotation in ipairs(tmp) do
	    CustomAnnotations.instantiate(annotation)
	end
	
	CustomAnnotations.pool = CustomAnnotations.calc_pool()
	CustomAnnotations.init_menus()
	
	for player in Players() do
	    player[CustomAnnotations.MENU] = CustomAnnotations.TOP_MENU
	    player[CustomAnnotations.REFRESH] = true
	    if not persist[-player.index - 1] then
		persist[-player.index - 1] = {
		    slots = {}
		}
	    end
	end
    end,

    idle = function()
	local player
	
	for player in Players() do
	    CustomAnnotations.detect_annotation(player)
	end

	CustomAnnotations.flash_editing_annotations()
    end,

    init_menus = function()
	CustomAnnotations.TOP_MENU = {
	    [2] = {
		prompt = CustomAnnotations.NEW,
		fn = CustomAnnotations.menu_new_annotation,
		is_fixed = true,
	    },
	    [3] = {
		prompt = CustomAnnotations.EDIT,
		fn = CustomAnnotations.menu_edit_annotation,
		is_fixed = true
	    },
	    [4] = {
		prompt = CustomAnnotations.DELETE,
		fn = CustomAnnotations.menu_delete_annotation,
		is_fixed = true,
	    },
	    [5] = {
		prompt = CustomAnnotations.SLOT_1,
		fn = CustomAnnotations.new_slot_1_annotation,
	    },
	    [6] = {
		prompt = CustomAnnotations.SLOT_2,
		fn = CustomAnnotations.new_slot_2_annotation,
	    },
	}

	CustomAnnotations.NEW_MENU = {
	    [2] = {
		prompt = CustomAnnotations.SLOT_1,
		fn = CustomAnnotations.new_slot_1_annotation,
	    },
	    [3] = {
		prompt = CustomAnnotations.SLOT_2,
		fn = CustomAnnotations.new_slot_2_annotation,
	    },
	    [4] = {
		prompt = CustomAnnotations.SLOT_3,
		fn = CustomAnnotations.new_slot_3_annotation,
	    },
	    [5] = {
		prompt = CustomAnnotations.SLOT_4,
		fn = CustomAnnotations.new_slot_4_annotation,
	    },
	    [6] = {
		prompt = CustomAnnotations.SLOT_5,
		fn = CustomAnnotations.new_slot_5_annotation,
	    },
            [7] = {
                prompt = CustomAnnotations.SLOT_6,
                fn = CustomAnnotations.new_slot_6_annotation,
            },
	}

	CustomAnnotations.EDIT_MENU = {
	    [2] = {
		prompt = CustomAnnotations.SLOT_1,
		fn = CustomAnnotations.edit_slot_1_annotation,
	    },
	    [3] = {
		prompt = CustomAnnotations.SLOT_2,
		fn = CustomAnnotations.edit_slot_2_annotation,
	    },
	    [4] = {
		prompt = CustomAnnotations.SLOT_3,
		fn = CustomAnnotations.edit_slot_3_annotation,
	    },
	    [5] = {
		prompt = CustomAnnotations.SLOT_4,
		fn = CustomAnnotations.edit_slot_4_annotation,
	    },
	    [6] = {
		prompt = CustomAnnotations.SLOT_5,
		fn = CustomAnnotations.edit_slot_5_annotation,
	    },
            [7] = {
                prompt = CustomAnnotations.SLOT_6,
                fn = CustomAnnotations.edit_slot_6_annotation,
            },
	}

	CustomAnnotations.DELETE_MENU = {
	    [2] = {
		prompt = CustomAnnotations.CONFIRM,
		fn = CustomAnnotations.menu_confirm_deletion,
		is_fixed = true,
	    },
	    [4] = {
		prompt = CustomAnnotations.CANCEL,
		fn = CustomAnnotations.menu_cancel_deletion,
		is_fixed = true,
	    },
	}
    end,

    detect_annotation = function(player)
	local hot, selection
        
	if player.has_map_open then
	    CustomAnnotations.show_menu(player)
	    hot = CustomLatches.get_hotkey(player)

	    if hot then
		selection = player[CustomAnnotations.MENU][hot]
		if selection then
		    selection.fn(player)
		end
	    end
	else
	    CustomAnnotations.hide_menu(player)
	end
    end,

    abbreviate = function(text)
        local abbreviation = CustomAnnotations.abbreviations[text]
        if not abbreviation then
            abbreviation = ""
            local i, prev, curr
            prev = " "
            for i = 1, string.len(text) do
                curr = string.sub(text, i, i)
                if prev == " " and curr ~= " " then
                    abbreviation = abbreviation .. curr
                end
                prev = curr
            end
            CustomAnnotations.abbreviations[text] = abbreviation
        end

        return abbreviation
    end,

    show_menu = function(player, hotkey)    
	local o, v, menu, config, overlay, prompt, is_fixed

	-- commented out in case some other script keeps blabbing
	-- if not player[CustomAnnotations.REFRESH] then return end
	
	o = 0
	menu = player[CustomAnnotations.MENU]
        for h = 1, 12 do
	    if o > 5 then break end
	    v = menu[h]
            
	    if v then
		prompt = v.prompt
		config = player.hotkey_bindings[h]
		overlay = player.overlays[o]
		is_fixed = v.is_fixed
                
		overlay:clear()

		if is_fixed then
		    if
			prompt == CustomAnnotations.DELETE
			or prompt == CustomAnnotations.CANCEL
		    then
			overlay.color = "red"
		    elseif prompt == CustomAnnotations.EDIT then
			overlay.color = "white"
		    else
			overlay.color = "green"
		    end
		else
		    prompt = CustomAnnotations.get_player_text(player, prompt)
		    overlay.color = "green"
		end
                
		overlay.text = CustomAnnotations.format_binding(
		    config,
		    prompt,
		    is_fixed
		)
                
		o = o + 1
	    end
	end

	while o < 6 do
	    overlay = player.overlays[o]
	    overlay:clear()
	    o = o + 1
	end

	player[CustomAnnotations.REFRESH] = false
    end,

    format_binding = function(config, action, fixed)
	local binding, len, missing

	if config.key then
	    binding = config.key
	elseif config.mouse then
	    binding = config.mouse
	elseif config.joystick then
	    binding = config.joystick
	else
	    binding = "<?>"
            missing = true
	end

        if not missing then
            binding = CustomAnnotations.abbreviate(binding)
        end

	if not fixed and string.len(action) > CustomAnnotations.CUTOFF then
	    action = string.sub(action, 1, CustomAnnotations.CUTOFF) ..
		CustomAnnotations.ELLIPSIS
	end
	
	return string.format("%s: %s", binding, action)
    end,

    hide_menu = function(player)
	local i
	
	CustomAnnotations.change_menu(
	    player,
	    CustomAnnotations.TOP_MENU
	)
	
	for i = 0, 5 do
	    player.overlays[i]:clear()
	end
    end,
    
    change_menu = function(player, menu)
	player[CustomAnnotations.MENU] = menu
	player[CustomAnnotations.REFRESH] = true
    end,

    calc_pool = function()
	local pool, a
        
	pool = {}
        
	for a in Annotations() do
	    if a.polygon == nil and a[CustomAnnotations.PERSIST] then
		table.insert(pool, a)
	    end
	end
	
	return pool
    end,

    remove_annotation = function(annotation)
	table.insert(CustomAnnotations.pool, annotation)
	annotation.polygon = nil
    end,

    create_annotation = function(polygon, text, x, y, player)
	local a
	
	a = table.remove(CustomAnnotations.pool)

	if player and #Players > 1 then
	    text = string.format("%s: %s", player.name, text)
	end

	if a then
	    a.polygon = polygon
	    a.text = text
	else
	    a = Annotations.new(
		polygon,
		text
	    )
	end
	
	a.x = x or polygon.x
	a.y = y or polygon.y

	CustomAnnotations.persist(a)

	return a
    end,

    persist = function(annotation)
	Game[CustomAnnotations.PERSIST][Level.index][annotation.index] =
	    CustomAnnotations.copy(annotation)
	annotation[CustomAnnotations.PERSIST] = true
    end,

    un_persist = function(annotation)
	Game[CustomAnnotations.PERSIST][Level.index][annotation.index] = nil
	annotations[CustomAnnotations.PERSIST] = false
    end,

    get_closest_annotation = function(player)
	local annotation, closest, distance, min

	min = CustomAnnotations.AUTO_RANGE + 0.001

	for annotation in Annotations() do
	    if annotation.polygon then
		distance = math.sqrt(
		    (player.x - annotation.x) ^ 2 +
		    (player.y - annotation.y) ^ 2
		)
		
		if distance < min then
		    closest = annotation
		    min = distance
		end
	    end
	end
	
	return closest
    end,


    menu_new_annotation = function(player)
	CustomAnnotations.change_menu(
	    player,
	    CustomAnnotations.NEW_MENU)
    end,

    menu_edit_annotation = function(player)
	local closest, menu
	
	closest = CustomAnnotations.get_closest_and_edit(player)

	if closest then
	    menu = CustomAnnotations.EDIT_MENU
	else
	    menu = CustomAnnotations.TOP_MENU
	end
	
	CustomAnnotations.change_menu(player, menu)
    end,

    menu_delete_annotation = function(player)
	local closest, menu
	
	closest = CustomAnnotations.get_closest_and_edit(player)

	if closest then
	    menu = CustomAnnotations.DELETE_MENU
	else
	    menu = CustomAnnotations.TOP_MENU
	end
	
	CustomAnnotations.change_menu(player, menu)
    end,

    get_closest_and_edit = function(player)
	local closest
	
	closest = CustomAnnotations.get_closest_annotation(player)

	if closest then
	    CustomAnnotations.start_editing(player, closest)
	end

	return closest
    end,

    menu_create_annotation = function(player, slot)
	local text

	text = CustomAnnotations.get_player_text(player, slot)
	
	CustomAnnotations.create_annotation(
	    player.polygon,
	    text,
	    player.x,
	    player.y,
	    player
	)
	
	CustomAnnotations.change_menu(
	    player,
	    CustomAnnotations.TOP_MENU
	)
    end,

    new_slot_1_annotation = function(player)
	CustomAnnotations.menu_create_annotation(
	    player,
	    CustomAnnotations.SLOT_1
	)
    end,
    
    new_slot_2_annotation = function(player)
	CustomAnnotations.menu_create_annotation(
	    player,
	    CustomAnnotations.SLOT_2
	)
    end,

    new_slot_3_annotation = function(player)
	CustomAnnotations.menu_create_annotation(
	    player,
	    CustomAnnotations.SLOT_3
	)
    end,

    new_slot_4_annotation = function(player)
	CustomAnnotations.menu_create_annotation(
	    player,
	    CustomAnnotations.SLOT_4
	)
    end,

    new_slot_5_annotation = function(player)
	CustomAnnotations.menu_create_annotation(
	    player,
	    CustomAnnotations.SLOT_5
	)
    end,

    new_slot_6_annotation = function(player)
	CustomAnnotations.menu_create_annotation(
	    player,
	    CustomAnnotations.SLOT_6
	)
    end,

    menu_change_annotation = function(player, slot)
	local annotation, text

	annotation = player[CustomAnnotations.EDITING]

	if annotation then
	    text = CustomAnnotations.get_player_text(player, slot)
	    annotation.text = text
	    annotation[CustomAnnotations.ORIGINAL] = nil
	    CustomAnnotations.stop_editing(player, annotation)
	    CustomAnnotations.persist(annotation)
	end
	
	CustomAnnotations.change_menu(
	    player,
	    CustomAnnotations.TOP_MENU
	)
    end,

    edit_slot_1_annotation = function(player)
	CustomAnnotations.menu_change_annotation(
	    player,
	    CustomAnnotations.SLOT_1
	)
    end,
    
    edit_slot_2_annotation = function(player)
	CustomAnnotations.menu_change_annotation(
	    player,
	    CustomAnnotations.SLOT_2
	)
    end,

    edit_slot_3_annotation = function(player)
	CustomAnnotations.menu_change_annotation(
	    player,
	    CustomAnnotations.SLOT_3
	)
    end,

    edit_slot_4_annotation = function(player)
	CustomAnnotations.menu_change_annotation(
	    player,
	    CustomAnnotations.SLOT_4
	)
    end,

    edit_slot_5_annotation = function(player)
	CustomAnnotations.menu_change_annotation(
	    player,
	    CustomAnnotations.SLOT_5
	)
    end,

    edit_slot_6_annotation = function(player)
	CustomAnnotations.menu_change_annotation(
	    player,
	    CustomAnnotations.SLOT_6
	)
    end,

    menu_confirm_deletion = function(player)
	local annotation

	annotation = player[CustomAnnotations.EDITING]
	
	if annotation then
	    CustomAnnotations.stop_editing(player, annotation)
	    CustomAnnotations.remove_annotation(annotation)
	    CustomAnnotations.persist(annotation)
	end
	
	CustomAnnotations.change_menu(
	    player,
	    CustomAnnotations.TOP_MENU
	)
    end,

    menu_cancel_deletion = function(player)
	local annotation

	annotation = player[CustomAnnotations.EDITING]
	
	if annotation then
	    CustomAnnotations.stop_editing(player, annotation)
	end
	
	CustomAnnotations.change_menu(
	    player,
	    CustomAnnotations.TOP_MENU
	)
    end,

    start_editing = function(player, annotation)
        local other, editing
        for other in Players() do
            if other ~= player then
                editing = other[CustomAnnotations.EDITING]
                if annotation == editing then return end
            end
        end
	player[CustomAnnotations.EDITING] = annotation
	CustomAnnotations.flashing[annotation.index] = true
    end,

    stop_editing = function(player, annotation)
	local original

	original = annotation[CustomAnnotations.ORIGINAL]

	if original then annotation.text = original end
	
	player[CustomAnnotations.EDITING] = nil
        annotation[CustomAnnotations.ORIGINAL] = nil
	CustomAnnotations.flashing[annotation.index] = false
    end,

    flash_editing_annotations = function()
	local idx, annotation, original

	for annotation in Annotations() do
	    if CustomAnnotations.flashing[annotation.index] then
		original = annotation[CustomAnnotations.ORIGINAL]

		if original == "" then
		    original = "<>"
		elseif not original then
		    original = annotation.text
		    annotation[CustomAnnotations.ORIGINAL] = original
		end

		if Game.ticks % 30 < 10 then
		    annotation.text = ""
		else
		    annotation.text = original
		end
	    end
	end
    end,

    copy = function(annotation)
	return {
	    index = annotation.index,
	    polygon = annotation.polygon,
	    x = annotation.x,
	    y = annotation.y,
	    text = annotation.text,
	}
    end,

    instantiate = function(annotation)
	local existing

	existing = Annotations[annotation.index]
	if annotation.polygon then
	    if existing then
		existing.polygon = annotation.polygon
		existing.x = annotation.x
		existing.y = annotation.y
		existing.text = annotation.text
	    else
		CustomAnnotations.create_annotation(
		    annotation.polygon,
		    annotation.text,
		    annotation.x,
		    annotation.y
		)
	    end
	end
    end,


    get_player_text = function(player, slot)
	local slots = Game[CustomAnnotations.PERSIST][-player.index - 1].slots
	return slots[slot] or CustomAnnotations.DEFAULTS[slot]
    end,

    set_player_text = function(player, slot, text)
	if text == "" then return end
	if not CustomAnnotations.DEFAULTS[slot] then return end
	local slots = Game[CustomAnnotations.PERSIST][-player.index - 1].slots
	slots[slot] = text
    end,

    clear_player_text = function(player, slot)
	CustomAnnotations.set_player_text(player, slot, nil)
    end,

    reset_player_text = function(player)
	local slot, _

	for slot, _ in ipairs(CustomAnnotations.DEFAULTS) do
	    CustomAnnotations.clear_player_text(player, slot)
	end
    end,
    
}

Triggers = {}

function Triggers.init(restoring_game)
    CustomPersistence.init(restoring_game)
    CustomLatches.init(restoring_game)
    CustomAnnotations.init(restoring_game)
end

function Triggers.idle()
    CustomLatches.idle()
    CustomAnnotations.idle()
end

-- solo shortcuts
presets = {
    set = function(slot, text)
        if #Players == 1 then
            CustomAnnotations.set_player_text(Players[0], slot, text)
        end
    end,

    clear = function(slot)
        if #Players == 1 then
            CustomAnnotations.clear_player_text(Players[0], slot)
        end
    end,

    reset = function(slot)
        if #Players == 1 then
            if slot then
                CustomAnnotations.clear_player_text(Players[0], slot)
            else
                CustomAnnotations.reset_player_text(Players[0])
            end
        end
    end,
    
}
